#include "emu.h"
#include "global.h"
#include <ddraw.h>
#include "input.h"
#include "gui.h"

#include "snes9x/port.h"
#include "snes9x/memmap.h"
#include "snes9x/display.h"
#include "snes9x/apu.h"
#include "snes9x/cpuexec.h"
#include "snes9x/soundux.h"
#include "snes9x/gfx.h"
#include "snes9x/debug.h"

bool enabled = false;

static LPDIRECTDRAWSURFACE4 screen = NULL;
static LPDIRECTDRAWSURFACE4 subscreen = NULL;

static unsigned short *GFXscreen = NULL;
static unsigned short *GFXsubscreen = NULL;
static unsigned char *GFXzbuffer = NULL;
static unsigned char *GFXsubzbuffer = NULL;

static FILE* td = NULL;

extern void S9xSyncSpeed() {
/*	static DWORD secondbegin = 0;
	static int framesthissec = 0;
	static int expectedticks = 0;
	if(!secondbegin) secondbegin = GetTickCount();
	
	int thistick = GetTickCount();
	if(thistick >= secondbegin + 1000) {
		secondbegin = thistick;
		framesthissec = 0;
		expectedticks = 0;
	}

	if(thistick > secondbegin + expectedticks + 10) {
		IPPU.RenderThisFrame = FALSE;
	} else {
		IPPU.RenderThisFrame = TRUE;
	}

	expectedticks += 18;
	framesthissec++; */
	IPPU.RenderThisFrame = TRUE;
}

//extern TCHAR* superstr;

void DoSnes9xInit(void) {
	ZeroMemory(&Settings, sizeof(Settings));

	Settings.SoundPlaybackRate = 4;
	Settings.Stereo = TRUE;
	Settings.SoundBufferSize = 0; 
	Settings.DisableSoundEcho = FALSE; 
	Settings.DisableSampleCaching = FALSE;
	Settings.DisableMasterVolume = FALSE;
	Settings.ThreadSound = FALSE;

	Settings.SPCTo65c816Ratio = 2;
	Settings.APUEnabled = TRUE;
	Settings.H_Max = (100 * SNES_CYCLES_PER_SCANLINE) / 100;
	Settings.SkipFrames = AUTO_FRAMERATE;
	Settings.Shutdown = TRUE;
	Settings.FrameTimePAL = 20000;
	Settings.FrameTimeNTSC = 16667;
	Settings.FrameTime = Settings.FrameTimeNTSC;
	Settings.Mouse = FALSE;
	Settings.SuperScope = FALSE; //TRUE;
	Settings.ControllerOption = SNES_MULTIPLAYER5;
	Settings.Transparency = FALSE; //TRUE;
	Settings.SixteenBit = TRUE;
	Settings.SupportHiRes = FALSE; //TRUE;
	Settings.NetPlay = FALSE;
	Settings.ServerName[0] = 0;
	Settings.AutoSaveDelay = 30;
//	Settings.TraceUnknownRegisters = TRUE;
//	Settings.DisplayFrameRate = TRUE;

	if(Settings.Transparency) Settings.SixteenBit = TRUE;

	Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

	if(!Memory.Init()){
		S9xMessage(0, 0, "Out of Memory, Memory.Init()");
	}

	if(!S9xInitAPU()) {
		S9xMessage(0, 0, "Out of Memory, S9xInitAPU()");
	}


/*#if defined UNDER_CE
	const char file[] = "\\CD-ROM\\CHRONO.SMC";
#else
	const char file[] = "CHRONO.SMC";
#endif

	if(!Memory.LoadROM(file)) {
#if defined USE_DIRECT_X
		n = wsprintf(str, TEXT("Failed to load ROM"));
		ExtTextOut(dc, 0, 32, NULL, NULL, str, n, NULL);
#endif
	} */

#if defined USE_DIRECT_X
	DDSURFACEDESC2			ddsd;
	HRESULT					hr;

	memset(&ddsd, 0, sizeof(ddsd));
	gdds_screen->GetSurfaceDesc(&ddsd);  

	ddsd.dwSize         = sizeof(DDSURFACEDESC2);
	ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_VIDEOMEMORY;

	ddsd.dwWidth        = 256;
	ddsd.dwHeight       = 239;
	hr = gpdd_directdraw->CreateSurface(&ddsd, &screen, NULL);
	if(hr != DD_OK) {
		S9xMessage(0, 0, "Failed to create DirectDraw screen surface");
	}

#if defined RENDER_DIRECT_TO_VIDEO_MEMORY
	hr = gpdd_directdraw->CreateSurface(&ddsd, &subscreen, NULL);

	if(hr != DD_OK) {
		S9xMessage(0, 0, "Failed to create DirectDraw subscreen surface");
	}
#endif
#endif

	GFXzbuffer = new unsigned char[256*239];
	GFXsubzbuffer = new unsigned char[256*239];
	GFX.Pitch = 256*2;
	GFX.ZBuffer = GFXzbuffer;
	GFX.SubZBuffer = GFXsubzbuffer;

#if !defined RENDER_DIRECT_TO_VIDEO_MEMORY
	GFXscreen = new unsigned short[256*239];
	GFXsubscreen = new unsigned short[256*239];

	GFX.Screen = (unsigned char *)GFXscreen;
	GFX.SubScreen = (unsigned char *)GFXsubscreen;
	S9xGraphicsInit();

	if(!GFXzbuffer || !GFXscreen || !GFXsubscreen || !GFXsubzbuffer) {
		S9xMessage(0, 0, "Out of Memory. Allocationg GFX buffers.");
	}
#endif

	S9xSetSoundMute(TRUE);

//	CPU.Flags |= DEBUG_MODE_FLAG;
}

void DoDisplayInputData(int dev, int data) {
#if defined USE_DIRECT_X
	static HFONT font = NULL;
	LOGFONT lf;
	if(!font) {
		memset(&lf, 0, sizeof(lf));
		lf.lfHeight = 32;
		font = CreateFontIndirect(&lf);
	}

	DDSURFACEDESC2	ddsd;
	DDCAPS			ddcaps;
	HDC				dc;
	TCHAR			str[1024];
	int n;

	ddsd.dwSize = sizeof(ddsd);
	gdds_screen->GetSurfaceDesc(&ddsd);
	gdds_screen->GetDC(&dc);

	SelectObject(dc, font);

	SetTextColor(dc, RGB(255, 255, 255));
	SetBkMode(dc, TRANSPARENT);

	TCHAR str2[1024];

	int line = 0;

	for(int i = 0; i < 12; i++) {
		if((data | BUTTON[i]) == data) {
			n = wsprintf(str, TEXT("Device %i, Button %i"), dev, i);
			ExtTextOut(dc, 0, line, NULL, NULL, str, n, NULL);
			line += 32;
		}
	}

	if((data | UP) == data) {
		n = wsprintf(str, TEXT("Device %i, UP"), dev, i);
		ExtTextOut(dc, 0, line, NULL, NULL, str, n, NULL);
		line += 32;
	}
	if((data | RIGHT) == data) {
		n = wsprintf(str, TEXT("Device %i, RIGHT"), dev, i);
		ExtTextOut(dc, 0, line, NULL, NULL, str, n, NULL);
		line += 32;
	}
	if((data | LEFT) == data) {
		n = wsprintf(str, TEXT("Device %i, LEFT"), dev, i);
		ExtTextOut(dc, 0, line, NULL, NULL, str, n, NULL);
		line += 32;
	}
	if((data | DOWN) == data) {
		n = wsprintf(str, TEXT("Device %i, DOWN"), dev, i);
		ExtTextOut(dc, 0, line, NULL, NULL, str, n, NULL);
		line += 32;
	}		

	gdds_screen->ReleaseDC(dc);
#endif

}


void DoSnes9xEmulation(void) {
	MSG msg;

	while(1) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		S9xMainLoop();

//		for(int i = 0; i < 4; i++) {
//			int j = GetInputDeviceData(i);
//			DoDisplayInputData(i, j);
//		}
			
	}
}

bool8 S9xInitUpdate () {
#if defined RENDER_DIRECT_TO_VIDEO_MEMORY
	DDSURFACEDESC2	ddsd;

	ddsd.dwSize = sizeof(ddsd);
	screen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
	GFX.Screen = (unsigned char *)ddsd.lpSurface;

	subscreen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
	GFX.SubScreen = (unsigned char *)ddsd.lpSurface;

	S9xGraphicsInit();
#endif

	return true; 
}

bool8 S9xDeinitUpdate (int width, int height, bool8 sixteen_bit) {
	static int frame = 0;

#if defined RENDER_DIRECT_TO_VIDEO_MEMORY
	screen->Unlock(NULL);
	subscreen->Unlock(NULL);
	S9xGraphicsDeinit();
#else
	DDSURFACEDESC2	ddsd;

	ddsd.dwSize = sizeof(ddsd);
	screen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);

	unsigned int *sp = (unsigned int *)ddsd.lpSurface;
	unsigned int *gp = (unsigned int *)GFXscreen;

	for(int y = 0; y < height; y++) {
		for(int x = 0; x < width/2; x++) {
//			if(x == 50) *(sp+x) = 0xFFFF;
//			else if(x == 100) *(sp+x) = frame % 63;
			/*else*/ *(sp+x) = *(gp+x);
		}
		sp += 128;
		gp += width/2;
	}

	screen->Unlock(NULL);
//	subscreen->Unlock(NULL);
#endif

	RECT rt = { 0, 0, width, height };

/*	TCHAR str[256];
	HDC dc;
	int n;
	screen->GetDC(&dc);
	SetTextColor(dc, RGB(255, 255, 255));
	SetBkMode(dc, TRANSPARENT);

	n = wsprintf(str, TEXT("%i x %i"), width, height);
	ExtTextOut(dc, 0, 0, NULL, NULL, str, n, NULL);

	screen->ReleaseDC(dc); */

	gdds_back->Blt(NULL, screen, &rt, DDBLT_WAIT, NULL);
	gdds_screen->Flip(NULL, DDFLIP_WAIT);

	frame++;
	return true; 
}

extern void S9xMessage(int type, int number, const char *message) {

#if !defined UNDER_CE
	static FILE *f = NULL;
	if(!f) {
		f = fopen("s9xmessage.trace.txt", "wt");
	}
	fprintf(f, "%s\n", message);
	fflush(f);
#endif

#if defined SHOW_S9XMESSAGE
#if defined USE_DIRECT_X
	static HFONT font = NULL;
	LOGFONT lf;
	if(!font) {
		memset(&lf, 0, sizeof(lf));
		lf.lfHeight = 32;
		font = CreateFontIndirect(&lf);
	}
	
	//	if(!enabled) return;

	DDBLTFX     ddbltfx;

    // Clear the back buffer
    memset(&ddbltfx, 0, sizeof(DDBLTFX));
    ddbltfx.dwSize = sizeof(DDBLTFX);
    ddbltfx.dwFillColor = 0;
    gdds_back->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);

	// Alright this code is here as overhead for showing text.
	// It's all over the place in my code so get over it
	// This code should not stay
	HDC				dc;
	TCHAR			str[1024];
	int n;

	gdds_back->GetDC(&dc);

	SelectObject(dc, font);

	SetTextColor(dc, RGB(255, 255, 255));
	SetBkMode(dc, TRANSPARENT);


#if defined UNDER_CE
	mbstowcs(str, message, strlen(message));
#else
	strcpy(str, message);
#endif

	n = strlen(message);
	ExtTextOut(dc, 0, 0, NULL, NULL, str, n, NULL);

	gdds_back->ReleaseDC(dc);
	gdds_screen->Flip(NULL, DDFLIP_WAIT);

//	int a = 0;
//	while(!a) {
//		for(int i = 0; i < 4; i++) {
//			a |= GetInputDeviceData(i);
//		}
//	}	

	MSG msg;

	DWORD tick = GetTickCount();
	while(GetTickCount() < tick + 1000) {
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
			if(!GetMessage(&msg, NULL, 0, 0))
				return;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

#endif
#endif
}

extern uint32 S9xReadJoypad (int which1_0_to_4) { 
	int n = which1_0_to_4;
	if(n == 4) return 0;
	int ret = 0;

#if !defined UNDER_CE
	n = 2;
#endif

	int data = GetInputDeviceData(n);

#if !defined UNDER_CE
	if((data | BUTTON[0]) == data) {
		ret |= SNES_A_MASK;
	}
	if((data | BUTTON[1]) == data) {
		ret |= SNES_B_MASK;
	}
	if((data | BUTTON[2]) == data) {
		ret |= SNES_X_MASK;
	}
	if((data | BUTTON[3]) == data) {
		ret |= SNES_Y_MASK;
	}
	if((data | BUTTON[4]) == data) {
		ret |= SNES_START_MASK;
	}
	if((data | BUTTON[5]) == data) {
		ret |= SNES_SELECT_MASK;
	}
	if((data | UP) == data) {
		ret |= SNES_UP_MASK;
	}
	if((data | DOWN) == data) {
		ret |= SNES_DOWN_MASK;
	}
	if((data | RIGHT) == data) {
		ret |= SNES_RIGHT_MASK;
	}
	if((data | LEFT) == data) {
		ret |= SNES_LEFT_MASK;
	}
	if((data | BUTTON[6]) == data) {
		ret |= SNES_TL_MASK;
	}
	if((data | BUTTON[7]) == data) {
		ret |= SNES_TR_MASK;
	}

	if((data | BUTTON[8]) == data) {
			DoInGameMenu();
	}

#else
	if((data | BUTTON[2]) == data) {
		if((data | BUTTON[1]) == data) {
			ret |= SNES_START_MASK;
		}
		if((data | BUTTON[8]) == data) {
			ret |= SNES_SELECT_MASK;
		}
		if(((data | BUTTON[9]) == data) && ((data | BUTTON[10]) == data)) {
			DoInGameMenu();
		}
	}

	if((data | BUTTON[0]) == data) {
		ret |= SNES_A_MASK;
	}
	if((data | BUTTON[1]) == data) {
		if((data | BUTTON[2]) != data) {
			ret |= SNES_B_MASK;
		}
	}
	if((data | BUTTON[7]) == data) {
		ret |= SNES_X_MASK;
	}
	if((data | BUTTON[8]) == data) {
		if((data | BUTTON[2]) != data) {
			ret |= SNES_Y_MASK;
		}
	}
	if((data | UP) == data) {
		ret |= SNES_UP_MASK;
	}
	if((data | DOWN) == data) {
		ret |= SNES_DOWN_MASK;
	}
	if((data | RIGHT) == data) {
		ret |= SNES_RIGHT_MASK;
	}
	if((data | LEFT) == data) {
		ret |= SNES_LEFT_MASK;
	}
	if((data | BUTTON[3]) == data) {
		ret |= SNES_UP_MASK;
	}
	if((data | BUTTON[4]) == data) {
		ret |= SNES_DOWN_MASK;
	}
	if((data | BUTTON[6]) == data) {
		ret |= SNES_RIGHT_MASK;
	}
	if((data | BUTTON[5]) == data) {
		ret |= SNES_LEFT_MASK;
	}
	if((data | BUTTON[10]) == data) {
		ret |= SNES_TL_MASK;
	}
	if((data | BUTTON[9]) == data) {
		ret |= SNES_TR_MASK;
	}
#endif

	ret |= 0x80000000;

	return ret;

}


